/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#include <string>
#include <sstream>

#include <liblot_exception.h>

#include "dkim-verify.h"
#include "dkim-config.h"

/* Get a configuration acceess object of the current method */
dkim_vrfy::config::cfg_access *dkim_vrfy::config::new_cfg_access()
{
	//return(new cfg_test);
	return(new cfg_xml);
}

void dkim_vrfy::config::cfg_test::read_cfg()
{
	parms_vals["mda"] = "/home/erik/dkim-verify/src/ioprint";
}

dkim_vrfy::config::cfg_xml::cfg_xml() : parser(0)
{
	// Initialize the XML4C2 system
	try {
		XMLPlatformUtils::Initialize();
	} catch (const XMLException &toCatch) {
		std::ostringstream msg;

		msg << "Error during Xerces-c Initialization.\n"
		<< "  Exception message:"
		<< StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;

		throw liblot::exception::app_error(msg.str());
	}

	parser = new XercesDOMParser;

	parser->setValidationScheme(XercesDOMParser::Val_Auto);
	parser->setDoNamespaces(true);
	parser->setDoSchema(false);
	parser->setValidationSchemaFullChecking(false);
	parser->setCreateEntityReferenceNodes(false);

	err_reporter = new DOMTreeErrorReporter();
	parser->setErrorHandler(err_reporter);
}

dkim_vrfy::config::cfg_xml::~cfg_xml()
{
    	delete err_reporter;
	err_reporter = 0;
	delete parser;
	parser = 0;

	XMLPlatformUtils::Terminate();
}

void dkim_vrfy::config::cfg_xml::read_cfg()
{
	using namespace dkim_vrfy::config;

	//int retval = 0;
    bool errorsOccurred = false;

    try {
        parser->parse(CFG_FILE);
    }
    catch (const OutOfMemoryException&)
    {
        XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
	errorsOccurred = true;
    }
    catch (const XMLException& e)
    {
        XERCES_STD_QUALIFIER cerr << "An error occurred during parsing\n   Message: "
             << StrX(e.getMessage()) << XERCES_STD_QUALIFIER endl;
	errorsOccurred = true;
    }

    catch (const DOMException& e)
    {
        const unsigned int maxChars = 2047;
        XMLCh errText[maxChars + 1];

        XERCES_STD_QUALIFIER cerr << "\nDOM Error during parsing: '"
		 << CFG_FILE << "'\n"
             << "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;

        if (DOMImplementation::loadDOMExceptionMsg(e.code, errText, maxChars))
             XERCES_STD_QUALIFIER cerr << "Message is: " << StrX(errText) << XERCES_STD_QUALIFIER endl;
	errorsOccurred = true;
    }

    catch (...)
    {
        XERCES_STD_QUALIFIER cerr << "An error occurred during parsing\n " << XERCES_STD_QUALIFIER endl;
	errorsOccurred = true;
    }


	    // If the parse was successful, output the document data from the DOM tree
    if (!errorsOccurred && !err_reporter->getSawErrors()) {
	        ConfigFilter cf;
		DOMDocument *doc = parser->getDocument();

	    DOMTreeWalker *tw = doc->createTreeWalker(doc,
						    DOMNodeFilter::SHOW_ALL,
						    &cf, true);
	if (tw == 0) {
		throw liblot::exception::app_error("Could not get treewalker");
	}

	DOMNode *node;
	for (node = tw->getRoot(); node; node = tw->nextNode()) {
		if (XMLString::equals(node->getNodeName(), element_mda)) {
			DOMNode *attr_node = node->getAttributes()->getNamedItem(attribute_value);

			std::ostringstream node_val;
				
			node_val << StrX(attr_node->getNodeValue());

			parms_vals[StrX(element_mda).localForm()] = node_val.str();
		}
	}
    }
}

bool DOMPrintErrorHandler::handleError(const DOMError &domError)
{
    // Display whatever error message passed from the serializer
    if (domError.getSeverity() == DOMError::DOM_SEVERITY_WARNING)
        XERCES_STD_QUALIFIER cerr << "\nWarning Message: ";
    else if (domError.getSeverity() == DOMError::DOM_SEVERITY_ERROR)
        XERCES_STD_QUALIFIER cerr << "\nError Message: ";
    else
        XERCES_STD_QUALIFIER cerr << "\nFatal Message: ";

    char *msg = XMLString::transcode(domError.getMessage());
    XERCES_STD_QUALIFIER cerr<< msg <<XERCES_STD_QUALIFIER endl;
    XMLString::release(&msg);

    // Instructs the serializer to continue serialization if possible.
    return true;
}

void DOMTreeErrorReporter::warning(const SAXParseException&)
{
    //
    // Ignore all warnings.
    //
}

void DOMTreeErrorReporter::error(const SAXParseException& toCatch)
{
    fSawErrors = true;
    XERCES_STD_QUALIFIER cerr << "Error at file \"" << StrX(toCatch.getSystemId())
                 << "\", line " << toCatch.getLineNumber()
                 << ", column " << toCatch.getColumnNumber()
         << "\n   Message: " << StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;
}

void DOMTreeErrorReporter::fatalError(const SAXParseException& toCatch)
{
    fSawErrors = true;
    XERCES_STD_QUALIFIER cerr << "Fatal Error at file \"" << StrX(toCatch.getSystemId())
                 << "\", line " << toCatch.getLineNumber()
                 << ", column " << toCatch.getColumnNumber()
         << "\n   Message: " << StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;
}

short ConfigFilter::acceptNode(const DOMNode* node) const
{
        //
        // The DOMWriter shall call getWhatToShow() before calling 
        // acceptNode(), to show nodes which are supposed to be 
        // shown to this filter.
        // 
        // REVISIT: In case the DOMWriter does not follow the protocol, 
        //          Shall the filter honour, or NOT, what it claimes
        //         (when it is constructed/setWhatToShow()) 
        //          it is interested in ?
        // 
        // The DOMLS specs does not specify that acceptNode() shall do
        // this way, or not, so it is up the implementation,
        // to skip the code below for the sake of performance ...
        //
        if ((getWhatToShow() & (1 << (node->getNodeType() - 1))) == 0)
                return DOMNodeFilter::FILTER_ACCEPT;

        switch (node->getNodeType())
        {
        case DOMNode::ELEMENT_NODE:
                {
                        if (XMLString::equals(node->getNodeName(), element_profile)) {
				DOMNode *attr_node = node->getAttributes()->getNamedItem(attribute_name);

				std::ostringstream node_val;
				
				node_val << StrX(attr_node->getNodeValue());
				
				if (node_val.str() != dkim_vrfy::main_app::instance()->get_profile()) {
                                	return DOMNodeFilter::FILTER_REJECT;
				}
			}
			
                        // for rest, accept it
                        return DOMNodeFilter::FILTER_ACCEPT;
                        break;
                }
        case DOMNode::COMMENT_NODE:
                {
                        // the WhatToShow will make this no effect
                        return DOMNodeFilter::FILTER_REJECT;
                        break;
                }
        case DOMNode::TEXT_NODE:
                {
                        // the WhatToShow will make this no effect
                        return DOMNodeFilter::FILTER_REJECT;
                        break;
                }
        case DOMNode::DOCUMENT_TYPE_NODE:
                {
                        // even we say we are going to process document type,
                        // we are not able be to see this node since
                        // DOMWriterImpl (a XercesC's default implementation
                        // of DOMWriter) will not pass DocumentType node to
                        // this filter.
                        //
                        return DOMNodeFilter::FILTER_REJECT;  // no effect
                        break;
                }
        case DOMNode::DOCUMENT_NODE:
                {
                        // same as DOCUMENT_NODE
                        return DOMNodeFilter::FILTER_REJECT;  // no effect
                        break;
                }
        default :
                {
                        return DOMNodeFilter::FILTER_ACCEPT;
                        break;
                }
        }

        return DOMNodeFilter::FILTER_ACCEPT;
}
