/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef DKIM_CONFIG_H
#define DKIM_CONFIG_H 1 

#include <iostream>
#include <string>
#include <map>

#include <xercesc/util/PlatformUtils.hpp>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMWriter.hpp>
#include <xercesc/dom/DOMWriterFilter.hpp>

#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>

#include <xercesc/util/OutOfMemoryException.hpp>

#include <xercesc/dom/DOMErrorHandler.hpp>

#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMError.hpp>

#include <xercesc/util/XercesDefs.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/ErrorHandler.hpp>

XERCES_CPP_NAMESPACE_USE

namespace dkim_vrfy {
	namespace config {
		class cfg_access;
		class cfg_xml;
		class cfg_test;
		cfg_access *new_cfg_access();
	}
}

class DOMPrintErrorHandler : public DOMErrorHandler
{
public:

    DOMPrintErrorHandler(){};
    ~DOMPrintErrorHandler(){};

    /** @name The error handler interface */
    bool handleError(const DOMError& domError);
    void resetErrors(){};

private :
    /* Unimplemented constructors and operators */
    DOMPrintErrorHandler(const DOMErrorHandler&);
    void operator=(const DOMErrorHandler&);

};

class DOMTreeErrorReporter : public ErrorHandler
{
public:
    // -----------------------------------------------------------------------
    //  Constructors and Destructor
    // -----------------------------------------------------------------------
    DOMTreeErrorReporter() :
       fSawErrors(false)
    {
    }

    ~DOMTreeErrorReporter()
    {
    }


    // -----------------------------------------------------------------------
    //  Implementation of the error handler interface
    // -----------------------------------------------------------------------
    void warning(const SAXParseException& toCatch);
    void error(const SAXParseException& toCatch);
    void fatalError(const SAXParseException& toCatch);
    void resetErrors() { fSawErrors = false; }

    // -----------------------------------------------------------------------
    //  Getter methods
    // -----------------------------------------------------------------------
    bool getSawErrors() const { return(fSawErrors); }

    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fSawErrors
    //      This is set if we get any errors, and is queryable via a getter
    //      method. Its used by the main code to suppress output if there are
    //      errors.
    // -----------------------------------------------------------------------
    bool    fSawErrors;
};

static const XMLCh  element_profile[]=
{
        chLatin_p, chLatin_r, chLatin_o, chLatin_f, chLatin_i, chLatin_l,
	chLatin_e, chNull
};

static const XMLCh  element_mda[]=
{
        chLatin_m, chLatin_d, chLatin_a, chNull
};

static const XMLCh  attribute_name[]=
{
        chLatin_n, chLatin_a, chLatin_m, chLatin_e, chNull
};

static const XMLCh  attribute_value[]=
{
        chLatin_v, chLatin_a, chLatin_l, chLatin_u, chLatin_e, chNull
};

class ConfigFilter : public DOMNodeFilter {
public:

    /** @name Constructors */
        ConfigFilter(unsigned long whatToShow = DOMNodeFilter::SHOW_ALL) :
		fWhatToShow(whatToShow) {}

    //@{

    /** @name Destructors */
        ~ConfigFilter(){};
    //@{

        /** @ interface from DOMWriterFilter */
        virtual short acceptNode(const DOMNode*) const;
    //@{

        virtual unsigned long getWhatToShow() const { return fWhatToShow; };

        virtual void          setWhatToShow(unsigned long toShow) { fWhatToShow = toShow; };

private:
        // unimplemented copy ctor and assignement operator
        ConfigFilter(const ConfigFilter&);
        ConfigFilter & operator = (const ConfigFilter&);

        unsigned long fWhatToShow;

};

// ---------------------------------------------------------------------------
//  This is a simple class that lets us do easy (though not terribly efficient)
//  trancoding of XMLCh data to local code page for display.
// ---------------------------------------------------------------------------
class StrX
{
public :
    // -----------------------------------------------------------------------
    //  Constructors and Destructor
    // -----------------------------------------------------------------------
    StrX(const XMLCh* const toTranscode)
    {
        // Call the private transcoding method
        fLocalForm = XMLString::transcode(toTranscode);
    }

    ~StrX()
    {
        XMLString::release(&fLocalForm);
    }


    // -----------------------------------------------------------------------
    //  Getter methods
    // -----------------------------------------------------------------------
    const char* localForm() const
    {
        return fLocalForm;
    }

private :
    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fLocalForm
    //      This is the local code page form of the string.
    // -----------------------------------------------------------------------
    char*   fLocalForm;
};

inline XERCES_STD_QUALIFIER ostream& operator<<(XERCES_STD_QUALIFIER ostream& target, const StrX& toDump)
{
    target << toDump.localForm();
    return target;
}

class dkim_vrfy::config::cfg_access {
public:
	virtual ~cfg_access() {}

	virtual void read_cfg() = 0;
	virtual std::map<std::string, std::string> &get_cfg() = 0;		
};

class dkim_vrfy::config::cfg_xml : public cfg_access {
	XercesDOMParser *parser;
	DOMTreeErrorReporter *err_reporter;
	std::map<std::string, std::string> parms_vals;
public:
	cfg_xml();
	~cfg_xml();

	void read_cfg();
	std::map<std::string, std::string> &get_cfg() {
		return(parms_vals);
	}
};

class dkim_vrfy::config::cfg_test : public cfg_access {
	std::map<std::string, std::string> parms_vals;
public:
	void read_cfg();
	std::map<std::string, std::string> &get_cfg() {
		return(parms_vals);
	}
};

#endif // DKIM_CONFIG_H
