/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <sstream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <vector>

#include <cstring>

extern "C" {
#include <dkim.h>
}

#include <liblot_exception.h>
#include <liblot_network.h>
#include <liblot_text.h>

#include "dkim-libdkim.h"
#include "dkim-rfc2822.h"

namespace dkim_vrfy {
	namespace libdkim {
		std::ostream &
		operator<<(std::ostream &os,
			   const dkim_vrfy::libdkim::dkim_helper &dh) throw()
		{
			os << dh.msg;

			return(os);
		}

		std::istream &
		operator>>(std::istream &is,
			   dkim_vrfy::libdkim::dkim_helper &dh) throw()
		{
			char c;
			std::ostringstream os(dh.msg);

			os.seekp(0, std::ios_base::end);
			while(is.get(c)) os.put(c);

			/* Store the raw message */
			dh.msg.assign(os.str());

			/* When the message arrives, we'll interpret
			 * it to save other methods time later on and
			 * to streamling error handling. */
			dh.msg_arrival_actions();

			return(is);
		}

		static const char * const result_lookup[] = {
			"none", // RESULT_NONE, 0
			"pass", // RESULT_PASS, 1
			"permerror", // RESULT_PERMERROR, 2
			"neutral", // RESULT_NEUTRAL, 3
		};
	}
}

const std::string
dkim_vrfy::libdkim::dkim_helper::AUTH_RESULT_HDR("Authentication-Results: ");
const char dkim_vrfy::libdkim::dkim_helper::AUTH_RESULT_DELIMITER = ';';
const std::string
dkim_vrfy::libdkim::dkim_helper::AUTH_RESULT_DKIM("dkim=");
const std::string
dkim_vrfy::libdkim::dkim_helper::AUTH_RESULT_DKIM_ASP("dkim-adsp=");
const std::string
dkim_vrfy::libdkim::dkim_helper::AUTH_RESULT_HEADER_I("header.i=");
const std::vector<std::string>
dkim_vrfy::libdkim::dkim_helper::result_dkim_lookup(result_lookup,
		result_lookup + sizeof(result_lookup)/sizeof(*result_lookup));

dkim_vrfy::libdkim::dkim_helper::dkim_helper(const std::string &id) :
		dkim(0), lib(0), eom_status(DKIM_STAT_OK),
        	has_from_(false), mailbox_split_ok(false),
		header_verify_ok(false), has_test_key(false),
		eolh(liblot::text::EOL_DELIM_LF)
{
	using namespace liblot::exception;

	DKIM_STAT status;
	static dkim_query_t qtype = DKIM_QUERY_DNS;	
	u_int opts = DKIM_LIBFLAGS_FIXCRLF;

        lib = dkim_init(0, 0);
	if (lib == 0) throw app_error("Unable to initialize libdkim");

        dkim_options(lib, DKIM_OP_SETOPT, DKIM_OPTS_QUERYMETHOD,
		     &qtype, sizeof qtype);

	dkim_options(lib, DKIM_OP_SETOPT, DKIM_OPTS_FLAGS,
		     &opts, sizeof opts);


        dkim = dkim_verify(lib, id.c_str(), 0, &status);
	if (dkim == 0) throw dkim_stat(status);
}

dkim_vrfy::libdkim::dkim_helper::~dkim_helper()
{
	dkim_free(dkim);
	dkim_close(lib);
}

void dkim_vrfy::libdkim::dkim_helper::msg_arrival_actions()
{
	using namespace dkim_vrfy::rfc2822;

	try {
		/* Set the message into msg_process for later use */
		msgp.set_msg(msg);
		has_from_ = true;
	} catch (msg_missing_from_ &e) {
		has_from_ = false;
	} catch (...) {} // We're not interested in other exceptions

	try {
		message_domain = get_message_domain();
		mailbox_split_ok = true;
	} catch (rfc2822_err &e) {
		mailbox_split_ok = false;	
	} catch (...) {} // we're not interested in other exceptions


	try {
		header_verify_ok =
			rfc2822_helper::
			reqd_header_verify(msgp.header_get_all());
	} catch (...) {} // we're not interested in exceptions
}

bool dkim_vrfy::libdkim::dkim_helper::authorsigok() const
{
        DKIM_STAT status;
        int nsigs;
        DKIM_SIGINFO **sigs;

	if (message_domain.empty()) return(false);

        status = dkim_getsiglist(dkim, &sigs, &nsigs);
        if (status != DKIM_STAT_OK) return(false);

        for (int c = 0; c < nsigs; c++) {
                if (dkim_sig_geterror(sigs[c]) != DKIM_SIGERROR_UNKNOWN &&
                    dkim_sig_geterror(sigs[c]) != DKIM_SIGERROR_OK) continue;

		char *sig_domain = dkim_sig_getdomain(sigs[c]);
		if (sig_domain == 0) return(false);

		if ((strcasecmp(dkim_sig_getdomain(sigs[c]),
                                message_domain.c_str()) == 0) &&
                    ((dkim_sig_getflags(sigs[c]) & DKIM_SIGFLAG_PASSED) != 0) &&
                    (dkim_sig_getbh(sigs[c]) == DKIM_SIGBH_MATCH)) return(true);
        }

        return(false);
}

dkim_vrfy::libdkim::dkim_helper::sig_status
dkim_vrfy::libdkim::dkim_helper::get_sig_status() const
{
	using namespace dkim_vrfy::rfc2822;

	/* Make sure e-mail has a "From " header; otherwise, BADFORMAT */
	if (! has_from_) return(STATUS_BADFORMAT);

	/* Make sure rfc2822_mailbox_split() succeeds; otherwise, BADFORMAT */
	if (! mailbox_split_ok) return(STATUS_BADFORMAT);

	/* Make sure message has required headers (see dkim-filter.c);
	 * otherwise, BADFORMAT */
	if (! header_verify_ok) return(STATUS_BADFORMAT);

	/* Check libdkim status for REVOKED, BADSIG, NOSIG, NOKEY, or OK */
	sig_status stat = STATUS_UNKNOWN;
	switch(eom_status) {
	case DKIM_STAT_REVOKED:
		stat = STATUS_REVOKED;	
		break;
	case DKIM_STAT_BADSIG:
		stat = STATUS_BAD;
		break;
	case DKIM_STAT_NOSIG:
		stat = STATUS_NOSIGNATURE;
		break;
	case DKIM_STAT_NOKEY:
		stat = STATUS_NOKEY;
		break;
	case DKIM_STAT_OK:
		stat = STATUS_GOOD;
		break;
	case DKIM_STAT_CANTVRFY:
		stat = STATUS_VERIFYERR;
		break;
	default:
		stat = STATUS_UNKNOWN;
	}

	return(stat);
}

std::string
dkim_vrfy::libdkim::dkim_helper::get_result_str(const dkim_sig &dks) const
{
	std::ostringstream result;

	std::string failstatus = (has_test_key ? "neutral" : "hardfail");

	switch (get_sig_status()) {
	case STATUS_GOOD:
		result << "pass";
		result << " " << "(" << dks.get_sig_keysize();
		result << "-bit key)";
		break;
	case STATUS_NOSIGNATURE:
		result << "none";
		result << " " << "(no signature)";
		break;
	case STATUS_BAD:
		result << failstatus;
		result << " " << "(verification failed)";
		break;
	case STATUS_REVOKED:
		result << failstatus;
		result << " " << "(revoked)";
		break;
	case STATUS_PARTIAL:
		result << "permerror";
		result << " " << "(partial verification)";
		break;
	case STATUS_VERIFYERR:
	{
		result << "permerror";
		result << " " << "(verification error";

		std::string err_str(get_error());

		if (! err_str.empty()) {
			result << ": " << err_str;
		}
		result << ")";
	}
		break;
	case STATUS_BADFORMAT:
		result << "permerror";
		result << " (bad format)";
		break;
	case STATUS_UNKNOWN:
	default:
		result << "neutral";
	}

	return(result.str());
}

std::string
dkim_vrfy::libdkim::dkim_helper::
auth_result_generate(const dkim_sig &dks) const
{
	using namespace dkim_vrfy::libdkim;

	std::ostringstream hdr;

	/* Logic/construction of header matches up with similar code
	 * in dkim-filter/dkim-filter.c from the dkim-milter package */
	hdr << AUTH_RESULT_HDR;
	hdr << liblot::network::hostname::get_hostname();
	hdr << AUTH_RESULT_DELIMITER;
	hdr << " " << AUTH_RESULT_DKIM << get_result_str(dks);
	hdr << " " << AUTH_RESULT_HEADER_I << dks.get_sig_identity();
	hdr << AUTH_RESULT_DELIMITER << " " << AUTH_RESULT_DKIM_ASP;
	hdr << get_policy_str();

	return(hdr.str());
}

std::string dkim_vrfy::libdkim::dkim_helper::get_message_domain() const
{
	using namespace liblot::exception;
	using namespace dkim_vrfy::rfc2822;

	std::vector<std::string> from_header(msgp.header_get("From:"));

	if (from_header.size() > 1) {
		throw app_error("Multiple \"From\" headers: can't determine domain");
	} else if (from_header.empty()) {
		throw app_error("No \"From\" header: can't determine domain");
	}

	header_parts h = rfc2822_helper::rfc2822_split(from_header.at(0));

	if (h.domain.empty()) throw app_error("Unable to retrieve domain");

	return(h.domain);
}

std::string dkim_vrfy::libdkim::dkim_helper::get_policy_domain() const
{
	char *domain = reinterpret_cast<char *>(dkim_getdomain(dkim));
	if (domain == 0) {
		liblot::exception::
			app_notice("Policy domain unavailable");
	}

	return(domain);
}

dkim_policy_t dkim_vrfy::libdkim::dkim_helper::get_policy() const
{
	using namespace liblot::exception;

	DKIM_STAT status;
	dkim_policy_t pcode;

	status = dkim_policy(dkim, &pcode, 0);

	if (status != DKIM_STAT_OK) {
		throw dkim_stat(status);
	}

	return(pcode);
}

std::string dkim_vrfy::libdkim::dkim_helper::get_policy_str() const
{
	using namespace liblot::exception;

	dkim_policy_t pcode;

	try {
		pcode = get_policy();
	} catch (...) {
		return("temperror");
	}

	if (get_presult() == DKIM_PRESULT_NXDOMAIN) {
		return("nxdomain");
	}

	std::string policy_str;
	bool authorsig = authorsigok();

	if (pcode == DKIM_POLICY_NONE) {
		policy_str = "none";
	} else if (((pcode == DKIM_POLICY_ALL) ||
		    (pcode == DKIM_POLICY_DISCARDABLE)) && authorsig) {
		policy_str = "pass";
	} else if (pcode == DKIM_POLICY_UNKNOWN) {
		if (authorsig) {
			policy_str = "signed";
		} else {
			policy_str = "unknown";
		}
	} else if ((pcode == DKIM_POLICY_ALL) && !authorsig) {
		policy_str = "fail";
	} else if ((pcode == DKIM_POLICY_DISCARDABLE) && !authorsig) {
		policy_str = "discard";
	}

	return(policy_str);
}

DKIM_STAT dkim_vrfy::libdkim::dkim_helper::chunk()
{
	using namespace liblot::exception;

	if (msg.empty()) throw app_error("No message");

	DKIM_STAT status;
	std::string tmp_msg;

	tmp_msg = liblot::text::eol_helper::nl_to_crlf(msg);

	char *buf = 0;
	int sz = tmp_msg.size();

	if (sz == 0) return(DKIM_STAT_INTERNAL);

	try {
		buf = new char[sz];
	} catch (std::bad_alloc &e) {
		std::cerr << e.what() << std::endl;
		return(DKIM_STAT_INTERNAL);
	}

	memcpy(buf, tmp_msg.data(), sz);
	
        status = dkim_chunk(dkim, reinterpret_cast<u_char *>(buf), sz);

	delete[] buf;

	return(status);
}

DKIM_STAT dkim_vrfy::libdkim::dkim_helper::eom()
{
        eom_status = dkim_chunk(dkim, 0, 0);
	if (eom_status == DKIM_STAT_OK) {
		/* TODO: The C/C++ bool types aren't compatible; research 
 		 * how to get a pointer to a C bool into C++ 
        	eom_status = dkim_eom(dkim, &has_test_key); */
        	eom_status = dkim_eom(dkim, 0);
	}

	return(eom_status);
}

/* Check to make sure we should be processing this message.  There are
 * two checks to be performed here:
 * 1. Check for a signature identity match (e.g. header.i in AR header
 * matches with the identity in a signature in the message)
 * 2. Was this AR header added by the same host as we're running on?
 * Both of these checks are true, then we won't look at the message.
 */
bool dkim_vrfy::libdkim::dkim_helper::check_handle_msg(const dkim_sig &dks)
{
	std::vector<std::string> ar_header(msgp.header_get(AUTH_RESULT_HDR));
	std::ostringstream os_identity;
	std::ostringstream os_host;
	bool host_match = false, identity_match = false;

	os_identity << AUTH_RESULT_HEADER_I << dks.get_sig_identity();
	os_host << AUTH_RESULT_HDR << liblot::network::hostname::get_hostname();

	for (std::vector<std::string>::const_iterator hi = ar_header.begin();
	     hi != ar_header.end(); ++hi) {
		size_t pos;

		pos = (*hi).find(os_identity.str());
		if (pos != std::string::npos) identity_match = true;

		pos = (*hi).find(os_host.str());
		if (pos == 0) host_match = true;
	}

	return((host_match && identity_match) ? false : true);
}

DKIM_STAT dkim_vrfy::libdkim::dkim_helper::msg_end()
{
	DKIM_STAT status;
	
	status = chunk();
	if (status == DKIM_STAT_OK) {
		status = eom();
	}

	return(status);
}

std::string dkim_vrfy::libdkim::dkim_helper::auth_result_insert(const dkim_sig &dks)
{
	std::string new_hdr = auth_result_generate(dks);

	msgp.header_insert_first(new_hdr);
	msg = msgp.get_msg();

	return(new_hdr);
}

const int dkim_vrfy::libdkim::dkim_sig::IDENTITY_MAX_SZ = 256;

dkim_vrfy::libdkim::dkim_sig::dkim_sig(const dkim_helper &d) :
			dkim(d.dkim), sig(dkim_getsignature(dkim)) {
		if (sig == 0) {
			throw liblot::exception::
				app_notice("No signature available");
		}
}

std::string dkim_vrfy::libdkim::dkim_sig::get_sig_domain() const
{
	char *domain = dkim_sig_getdomain(sig);

	if (domain == 0) {
		liblot::exception::
			app_notice("Signature domain unavailable");
	}

	return(domain);
}

std::string dkim_vrfy::libdkim::dkim_sig::get_sig_identity() const
{
	std::string id_ret;
	char identity[IDENTITY_MAX_SZ];

	/* dkim_sig_getidentity() uses snprintf() or dkim_qp_decode()
 	 * to populate the buffer; if dkim_qp_decode() is used, the
 	 * string isn't NULL-terminated.  Clearing the buffer beforehand
 	 * prevents an overrun */
	bzero(identity, IDENTITY_MAX_SZ);

	if (dkim_sig_getidentity(dkim, sig, identity,
				 IDENTITY_MAX_SZ) == DKIM_STAT_OK) {
		id_ret.assign(identity);
	}

	return(id_ret);
}

unsigned int dkim_vrfy::libdkim::dkim_sig::get_sig_keysize() const
{
	unsigned int keybits = 0;

	dkim_sig_getkeysize(sig, &keybits);
	
	return(keybits);
}

dkim_vrfy::libdkim::dkim_sigerr::dkim_sigerr(const dkim_sig &s) :
			sigerr(DKIM_SIGERROR_UNKNOWN) {
	if ((s.dkim == 0) || (s.sig == 0)) {
		throw liblot::exception::
			app_error("No signature available");
	}

	sigerr = dkim_sig_geterror(s.sig);
}

namespace dkim_vrfy {
	namespace libdkim {
		static const char * const stat_lookup[] = {
			"OK", // DKIM_STAT_OK, 0
			"Bad signature", // DKIM_STAT_BADSIG, 1
			"No signature available", // DKIM_STAT_NOSIG, 2
			"No public key available", // DKIM_STAT_NOKEY, 3
			"Can't get domain key", // DKIM_STAT_CANTVRFY, 4
			"Message syntax invalid", // DKIM_STAT_SYNTAX, 5
			"Resource unavailable", // DKIM_STAT_NORESOURCE, 6
			"Internal error", // DKIM_STAT_INTERNAL, 7
			"Key revoked", // DKIM_STAT_REVOKED, 8
			"Invalid function parameter", // DKIM_STAT_INVALID, 9
			// DKIM_STAT_NOTIMPLEMENT, 10
			"Function not implemented",
			"Key retrieval failed", // DKIM_STAT_KEYFAIL, 11
			"Callback requested reject", // DKIM_STAT_CBREJECT, 12
			// DKIM_STAT_CBINVALID, 13
			"Callback gave invalid result",
			// DKIM_STAT_CBTRYAGAIN, 14
			"Callback says try again later",
			"Callback error", // DKIM_STAT_CBERROR, 15
			"Multiple DNS replies", // DKIM_STAT_MULTIDNSREPLY, 16
		};
	}
}

const std::vector<const char *>
dkim_vrfy::libdkim::dkim_stat::dkim_stat_lookup(stat_lookup,
		stat_lookup + sizeof(stat_lookup)/sizeof(*stat_lookup));
