/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef DKIM_LIBDKIM_H
#define DKIM_LIBDKIM_H 1

#include <iostream>
#include <string>
#include <vector>

#include <dkim.h>

#include <liblot_text.h>
#include <liblot_exception.h>

#include "dkim-rfc2822.h"

namespace dkim_vrfy {
	namespace libdkim {
		class dkim_helper;
		class dkim_sig;
		class dkim_stat;
		class dkim_sigerr;
	}
}

class dkim_vrfy::libdkim::dkim_helper {
	static const std::string AUTH_RESULT_HDR;
	static const char AUTH_RESULT_DELIMITER;
	static const std::string AUTH_RESULT_DKIM;
	static const std::string AUTH_RESULT_DKIM_ASP;
	static const std::string AUTH_RESULT_HEADER_I;
	static const std::vector<std::string> result_dkim_lookup;

	enum sig_status {
		STATUS_GOOD,
		STATUS_BAD,
		STATUS_NOKEY,
		STATUS_REVOKED,
		STATUS_NOSIGNATURE,
		STATUS_BADFORMAT,
		STATUS_PARTIAL,
		STATUS_VERIFYERR,
		STATUS_UNKNOWN,
	};

	enum result_dkim {
		RESULT_NONE = 0,
		RESULT_PASS,
		RESULT_PERMERROR,
		RESULT_NEUTRAL,
	};

        DKIM *dkim;
        DKIM_LIB *lib;
	DKIM_STAT eom_status;
	bool has_from_;
	bool mailbox_split_ok;
	bool header_verify_ok;
	bool has_test_key;
	std::string message_domain;

	mutable std::string msg;
	liblot::text::eol_helper eolh;
	dkim_vrfy::rfc2822::msg_process msgp;

	void msg_arrival_actions();

	bool authorsigok() const;
	sig_status get_sig_status() const;
	std::string get_result_str(const dkim_sig &) const;
	std::string get_policy_str() const;
	std::string get_message_domain() const;
	std::string auth_result_generate(const dkim_sig &) const;
	std::string get_policy_domain() const;
	dkim_policy_t get_policy() const;
	int get_presult() const { return(dkim_getpresult(dkim)); }
	std::string get_error() const {
		const char *err = dkim_geterror(dkim);

		return(err ? err : "");
	}

	/* These functions operate on msg */
	DKIM_STAT chunk();
	DKIM_STAT eom();

	/* Don't allow copies */
	dkim_helper(const dkim_helper &);
	dkim_helper &operator=(const dkim_helper &);
public:
	dkim_helper(const std::string &id = "libdkim");
	~dkim_helper();

	bool check_handle_msg(const dkim_sig &);
	DKIM_STAT msg_end();
	std::string auth_result_insert(const dkim_sig &);
	std::string get_msg() const { return(msg); }

	friend std::ostream &
	operator<<(std::ostream &, const dkim_helper &) throw();
	friend std::istream &
	operator>>(std::istream &, dkim_helper &) throw();

	friend class dkim_sig;
};

class dkim_vrfy::libdkim::dkim_sig {
	static const int IDENTITY_MAX_SZ;

	DKIM *dkim;
	DKIM_SIGINFO *sig;

	/* Don't allow copies */
	dkim_sig(const dkim_sig &);
	dkim_sig &operator=(const dkim_sig &);
public:
	dkim_sig() : dkim(0), sig(0) {}
	dkim_sig(const dkim_helper &);

	void set_dkim(DKIM *d) { dkim = d; }
	void set_sig(DKIM_SIGINFO *s) { sig = s; }

	std::string get_sig_domain() const;
	std::string get_sig_identity() const;
	unsigned int get_sig_keysize() const;
	unsigned int get_sig_getflags() const {
		return(dkim_sig_getflags(sig));
	}
	int get_sig_getbh() const {
		/* A possible value, DKIM_SIGBH_UNTESTED, can be -1,
		 * so this can't be an unsigned int in C++; the libdkim
		 * C API declares dkim_sig_getbh() as returning an
		 * unsigned int. */
		return(static_cast<int>(dkim_sig_getbh(sig)));
	}

	friend class dkim_sigerr;
};

class dkim_vrfy::libdkim::dkim_sigerr :
		public liblot::exception::app_exception {
	DKIM_SIGERROR sigerr;
public:
	dkim_sigerr(const dkim_sig &s);
	virtual ~dkim_sigerr() throw () {}

	DKIM_SIGERROR get_sigerr() { return(sigerr); }

	virtual const char *what() const throw() {
		return(dkim_sig_geterrorstr(sigerr));
	}
	std::string get_str() const { return(what()); }
	std::string get_id() const { return("DKIM Signature Error"); }
};

class dkim_vrfy::libdkim::dkim_stat :
		public liblot::exception::app_exception {
	DKIM_STAT stat;
	static const std::vector<const char *> dkim_stat_lookup;
public:
	dkim_stat(DKIM_STAT e) : stat(e) {} 
	virtual ~dkim_stat() throw () {}

	DKIM_STAT get_stat() { return(stat); }

	virtual const char *what() const throw() {
		return(dkim_stat_lookup.at(stat));
	}
	std::string get_str() const { return(what()); }
	std::string get_id() const { return("DKIM Status"); }
};

#endif // DKIM_LIBDKIM_H
