/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <sstream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <vector>
#include <list>

#include <cstring>

extern "C" {
#include <dkim.h>
}

#include <liblot_exception.h>

#include "dkim-rfc2822.h"

const int dkim_vrfy::rfc2822::msg_process::LINE_BREAK = 75;
const std::string dkim_vrfy::rfc2822::msg_process::WS_CHARS(" \t");
const std::string dkim_vrfy::rfc2822::msg_process::HEADER_NAME_DELIM(":");
const int dkim_vrfy::rfc2822::msg_process::msg_eol::HDR_BODY_SEP_EOL_NUM = 2;

void dkim_vrfy::rfc2822::msg_process::parse_header(const std::string &headers)
{
	std::istringstream is(eolh.to_nl(headers));
	std::string current_header;

	while (is.good()) {
		std::string header_line;

		std::getline(is, header_line);

		msg.headers.push_back(header_line);
	}
}

void dkim_vrfy::rfc2822::msg_process::msg_parse(const std::string &m)
{
	using namespace liblot::exception;

	static const std::string from_("From ");
	size_t hdr_sep_pos = m.find(eolh.get_body_delim());
	size_t from_end_pos;

	if (has_from_) {
		/* Grab "From " header */
		size_t pos = m.find(from_);
		if (pos == std::string::npos) {
			throw msg_missing_from_();
		}
		/* "From " header won't be folded */
		from_end_pos = m.find(eolh.get_delim(), pos);
		msg.from = m.substr(pos, from_end_pos);

		/* Skip EOL delim */
		from_end_pos += eolh.get_delim().size();
	} else {
		from_end_pos = 0;
	}

	/* Parse Headers into map */
	parse_header(m.substr(from_end_pos, hdr_sep_pos - from_end_pos));

	/* Message body */
	msg.body = m.substr(hdr_sep_pos + msg_eol::HDR_BODY_SEP_EOL_NUM);
}

std::string dkim_vrfy::rfc2822::msg_process::msg_unparse() const
{
	std::ostringstream os;

	if (has_from_) {
		/* "From " header/message separator */
		os << msg.from << eolh.get_delim();
	}

	/* Headers */
	typedef std::list<std::string>::const_iterator CI;
	for (CI mi = msg.headers.begin(); mi != msg.headers.end(); ++mi) {
		os << *mi << eolh.get_delim();
	}

	/* Message body */
	os << eolh.get_delim();
	os << msg.body;

	return(os.str());
}

void
dkim_vrfy::rfc2822::msg_process::header_insert_first(const std::string &header)
{
	using namespace liblot::exception;

	if (header.empty()) throw app_error("Header not specified");

	msg.headers.push_front(header);
}

void
dkim_vrfy::rfc2822::msg_process::header_insert_after(const std::string &header,
						     const std::string &needle)
{
	using namespace liblot::exception;

	if (header.empty()) throw app_error("Header not specified");
	if (needle.empty()) throw app_error("Needle not specified");

	typedef std::list<std::string>::iterator I;
	for (I li = msg.headers.begin(); li != msg.headers.end(); ++li) {
		size_t pos = (*li).find(needle);

		if (pos == 0) {
			msg.headers.insert(++li, header);
			break;
		}
	}
}

void
dkim_vrfy::rfc2822::msg_process::header_insert_before(const std::string &header,
						     const std::string &needle)
{
	using namespace liblot::exception;

	if (header.empty()) throw app_error("Header not specified");
	if (needle.empty()) throw app_error("Needle not specified");

	typedef std::list<std::string>::iterator I;
	for (I li = msg.headers.begin(); li != msg.headers.end(); ++li) {
		size_t pos = (*li).find(needle);

		if (pos == 0) {
			msg.headers.insert(li, header);
			break;
		}
	}
}

void
dkim_vrfy::rfc2822::msg_process::header_insert_last(const std::string &header)
{
	using namespace liblot::exception;

	if (header.empty()) throw app_error("Header not specified");

	msg.headers.push_back(header);
}

void dkim_vrfy::rfc2822::msg_process::fold(std::string &str) const
{
	size_t pos = str.find_last_of(WS_CHARS, LINE_BREAK);

	str.insert(pos, eolh.get_delim());
}

size_t dkim_vrfy::rfc2822::msg_process::unfold_at(std::string &str,
						  size_t spos) const
{	
	size_t ws_pos = std::string::npos;

	size_t pos = str.find(eolh.get_delim(), spos);
	if (pos != std::string::npos) {
		ws_pos = str.find_first_of(WS_CHARS, pos);
		if (ws_pos != std::string::npos) {
			if ((ws_pos - pos) == 1) { // next char WS
				str.erase(pos, eolh.get_delim().size());
			}
		}
	}

	return(ws_pos);
}

/* Folds each header in the message */
std::string
dkim_vrfy::rfc2822::msg_process::
header_fold_all(const std::string &headers) const
{
	std::istringstream is(eolh.to_nl(headers));
	std::ostringstream os;

	while (is.good()) {
		std::string header_line;

		std::getline(is, header_line);
		fold(header_line);

		os << header_line << eolh.get_delim();
	}

	return(os.str());
}

/* Unfolds each header in the message */
std::string
dkim_vrfy::rfc2822::msg_process::
		header_unfold_all(const std::string &headers) const
{
	std::string unfolded_headers(headers);
	size_t i = 0;

	while (i != std::string::npos) {
		i = unfold_at(unfolded_headers, i);
	}

	return(unfolded_headers);
}

/* Folds a single header line
 * Fold at specified width; break with "\n\t" */
std::string
dkim_vrfy::rfc2822::msg_process::header_fold(const std::string &header) const
{
	std::string folded_header(header);

	fold(folded_header);

	return(folded_header);
}

/* Unfolds a single header line
 * Unfolding is implemented by removing the EOL delimiter that precedes WS */
std::string
dkim_vrfy::rfc2822::msg_process::
		header_unfold(const std::string &header) const
{
	std::string unfolded_header(header);

	unfold_at(unfolded_header, 0);	

	return(unfolded_header);
}

std::vector<std::string>
dkim_vrfy::rfc2822::msg_process::
		header_get(const std::string &header_name) const
{
	std::vector<std::string> matching_headers;

	typedef std::list<std::string>::const_iterator CI;
	for (CI mi = msg.headers.begin(); mi != msg.headers.end(); ++mi) {
		size_t pos = (*mi).find(header_name);

		if (pos == 0) {
			std::string found_header(*mi);
		
			/* We found a match, but the header might be folded */
			while ((*(++mi)).find_first_of(WS_CHARS) == 0) {
				found_header.append(*mi);	
			}

			matching_headers.push_back(found_header);
			break;
		}
	}

	return(matching_headers);
}

dkim_vrfy::rfc2822::header_parts
dkim_vrfy::rfc2822::rfc2822_helper::
rfc2822_split(const std::string &header)
{
	using namespace dkim_vrfy::rfc2822;
	using namespace liblot::exception;

	char *line;
	char *user, *domain;
	int ret;
	bool errors = false;
	header_parts hp;

	line = strdup(header.c_str());

	ret = rfc2822_mailbox_split(line, &user, &domain);
	if (ret != RFC2822_OK) {
		errors = true;
	} else {
		if (user) hp.user.assign(user);
		if (domain) hp.domain.assign(domain);
	}

	free(line);

	if (errors) throw rfc2822_err(ret);

	return(hp);
}

bool
dkim_vrfy::rfc2822::rfc2822_helper::
reqd_header_verify(const std::list<std::string> &headers)
{
#if 0
                bool ok = TRUE;

                /* exactly one From: */
                if (dkimf_findheader(dfc, "From", 0) == NULL ||
                    dkimf_findheader(dfc, "From", 1) != NULL)
                        ok = FALSE;

                /* exactly one Date: */
                if (dkimf_findheader(dfc, "Date", 0) == NULL ||
                    dkimf_findheader(dfc, "Date", 1) != NULL)
                        ok = FALSE;

                /* no more than one Reply-To: */
                if (dkimf_findheader(dfc, "Reply-To", 1) != NULL)
                        ok = FALSE;

                /* no more than one To: */
                if (dkimf_findheader(dfc, "To", 1) != NULL)
                        ok = FALSE;

                /* no more than one Cc: */
                if (dkimf_findheader(dfc, "Cc", 1) != NULL)
                        ok = FALSE;

                /* no more than one Bcc: */
                if (dkimf_findheader(dfc, "Bcc", 1) != NULL)
                        ok = FALSE;

                /* no more than one Message-Id: */
                if (dkimf_findheader(dfc, "Message-Id", 1) != NULL)
                        ok = FALSE;

                /* no more than one In-Reply-To: */
                if (dkimf_findheader(dfc, "In-Reply-To", 1) != NULL)
                        ok = FALSE;

                /* no more than one References: */
                if (dkimf_findheader(dfc, "References", 1) != NULL)
                        ok = FALSE;

                /* no more than one Subject: */
                if (dkimf_findheader(dfc, "Subject", 1) != NULL)
                        ok = FALSE;
                if (!ok)
                {
                        if (conf->conf_dolog)
                        {
                                syslog(LOG_INFO,
                                       "%s RFC2822 header requirement error",
                                       dfc->mctx_jobid);
                        }

                        dfc->mctx_addheader = TRUE;
                        dfc->mctx_headeronly = TRUE;
                        dfc->mctx_status = DKIMF_STATUS_BADFORMAT;
                        return SMFIS_CONTINUE;
                }

#endif
	return(true);
}

namespace dkim_vrfy {
	namespace rfc2822 {
		static const char * const err_lookup[] = {
			"OK", // RFC2822_OK, 0
			"Unbalanced parentheses", // RFC2822_ERR_PUNBALANCED, 1
			"Unbalanced quotes", // RFC2822_ERR_QUNBALANCED, 2
			// RFC2822_ERR_SUNBALANCED, 3
			"Unbalanced square brackets",
		};
	}
}

const std::vector<const char *>
dkim_vrfy::rfc2822::rfc2822_err::rfc2822_err_lookup(err_lookup,
		   err_lookup + sizeof(err_lookup)/sizeof(*err_lookup));
