/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef DKIM_RFC2822_H
#define DKIM_RFC2822_H 1

#include <iostream>
#include <string>
#include <list>
#include <vector>

#include <dkim.h>

#include <liblot_exception.h>
#include <liblot_text.h>

namespace dkim_vrfy {
	namespace rfc2822 {
		/* The rfc2822 API isn't public so we define these
		 * ourselves; this should be in sync with rfc2822.c */
		enum rfc2822_err_code {
			RFC2822_OK = 0, // success
			RFC2822_ERR_PUNBALANCED = 1, // unbalanced paren.
			RFC2822_ERR_QUNBALANCED = 2, // unbalanced quotes
			RFC2822_ERR_SUNBALANCED = 3, // unbal. sq. brackets
		};

		struct rfc2822_msg;
		class msg_process;
		class header_empty;
		class header_not_found;
		class msg_format_invalid;
		class msg_missing_from_;
		struct header_parts;
		class rfc2822_helper;
		class rfc2822_err;
	}
}

struct dkim_vrfy::rfc2822::rfc2822_msg {
	std::string from;
	std::list<std::string> headers;
	std::string body;
};

/* This class contains methods to process RFC2822 e-mail messages including
 * inserting message headers, getting a list of headers for the application
 * to process, etc. */
class dkim_vrfy::rfc2822::msg_process {
	static const int LINE_BREAK;
	static const std::string WS_CHARS;
	static const std::string HEADER_NAME_DELIM;

	class msg_eol : public liblot::text::eol_helper {
	public:
		/* Number of EOL delimiters separating header from body */
		static const int HDR_BODY_SEP_EOL_NUM;

		msg_eol(liblot::text::eol_delim_style e) :
				eol_helper(e) {}

		std::string get_body_delim() const {
			return(eol_helper::get_delim(HDR_BODY_SEP_EOL_NUM));
		}
	};

	msg_eol eolh;
	rfc2822_msg msg;
	bool has_from_;

	void fold(std::string &) const;
	size_t unfold_at(std::string &, size_t) const;
	void parse_header(const std::string &);
	void msg_parse(const std::string &);
	std::string msg_unparse() const;
public:
	msg_process(liblot::text::eol_delim_style e =
		    liblot::text::EOL_DELIM_LF,
		    bool fr = false) : eolh(e), has_from_(fr) {}
	msg_process(const std::string &m,
		    liblot::text::eol_delim_style e =
		    liblot::text::EOL_DELIM_LF,
		    bool fr = false) : eolh(e), has_from_(fr) {
		msg_parse(m);
	}

	void set_msg(const std::string &m) {
		msg_parse(m);
	}
	std::string get_msg() const { return(msg_unparse()); }

	void header_insert_first(const std::string &);
	void header_insert_after(const std::string &, const std::string &);
	void header_insert_before(const std::string &, const std::string &);
	void header_insert_last(const std::string &);

	std::string header_fold_all(const std::string &) const;
	std::string header_unfold_all(const std::string &) const;
	std::string header_fold(const std::string &) const;
	std::string header_unfold(const std::string &) const;
	std::vector<std::string> header_get(const std::string &) const;
	std::list<std::string> header_get_all() const {
		return(msg.headers);	
	}
};

class dkim_vrfy::rfc2822::header_empty :
	public liblot::exception::app_notice {
public:
	header_empty(const std::string &s) : app_notice(s) {}
};

class dkim_vrfy::rfc2822::header_not_found :
	public liblot::exception::app_notice {
public:
	header_not_found(const std::string &s) : app_notice(s) {}
};

class dkim_vrfy::rfc2822::msg_format_invalid :
	public liblot::exception::app_error {
public:
	msg_format_invalid(const std::string &s) : app_error(s) {}
};

class dkim_vrfy::rfc2822::msg_missing_from_ : public msg_format_invalid {
public:
	msg_missing_from_() : msg_format_invalid("Missing \"From \" header") {}
};

struct dkim_vrfy::rfc2822::header_parts {
	std::string user;
	std::string domain;
};

/* This is a utility class providing functionality directly related
 * to the RFC2822 specification. */
class dkim_vrfy::rfc2822::rfc2822_helper {
public:
	static header_parts rfc2822_split(const std::string &);
	static bool reqd_header_verify(const std::list<std::string> &);
};

class dkim_vrfy::rfc2822::rfc2822_err :
		public liblot::exception::app_exception {
	static const std::vector<const char *> rfc2822_err_lookup;
	int err;
public:
	rfc2822_err(int e) : err(e) {} 
	virtual ~rfc2822_err() throw () {}

	int get_err() { return(err); }

	virtual const char *what() const throw() {
		return(rfc2822_err_lookup.at(err));
	}
	std::string get_str() const { return(what()); }
	std::string get_id() const { return("RFC2822 Error"); }
};

#endif // DKIM_RFC2822_H
