/* -*- c++ -*- */
/* 
 * This file is part of dkim-verify
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <iostream>
#include <string>
#include <map>
#include <sstream>
#include <list>
#include <exception>

#include <cstdio>  // perror()
#include <cstdlib> // atexit()
#include <csignal>

#include <pthread.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <liblot_exception.h>
#include <liblot_logging.h>
#include <liblot_cmd.h>

#include "dkim-verify.h"
#include "dkim-config.h"
#include "dkim-libdkim.h"

#include <liblot_args.h>

const std::string dkim_vrfy::main_app::SYSLOG_NAME("dkim-verify");
const std::string dkim_vrfy::main_app::VERSION("0.6.2");
const std::string dkim_vrfy::main_app::COPYRIGHT_YEAR("2008");
const std::string dkim_vrfy::main_app::AUTHOR("Erik Lotspeich");
const std::string dkim_vrfy::main_app::AUTHOR_EMAIL("erik@lotspeich.org");
const std::string dkim_vrfy::main_app::LICENSE("GPL v2");

namespace dkim_vrfy {
	int quit_th_semid = ERROR;
	pthread_t th;
	int main_ret = 0;

	void exit_main(int ret) {
		main_ret = ret;

		kill(getpid(), SIGTERM);		
		pthread_join(th, 0);
	
		_exit(ret);
	}
}

extern "C" void dkim_vrfy_atexit();
extern "C" void *quit_th(void *);

extern "C" void dkim_vrfy_atexit()
{
	using namespace dkim_vrfy;

	delete dkim_vrfy::main_app::instance();

	union semun arg;
	if (semctl(quit_th_semid, 0, IPC_RMID, arg) < 0) {
		perror("dkim_vrfy_atexit semctl");
	}
}

extern "C" void *quit_th(void *arg)
{
	using namespace dkim_vrfy;

	main_app *mai = static_cast<main_app *>(arg);
        int sig_caught;
	sigset_t sigs_towatch;

	sigemptyset(&sigs_towatch);
	sigaddset(&sigs_towatch, SIGABRT);
	sigaddset(&sigs_towatch, SIGTERM);
	sigaddset(&sigs_towatch, SIGQUIT);
	sigaddset(&sigs_towatch, SIGINT);
	sigaddset(&sigs_towatch, SIGUSR1);
	sigaddset(&sigs_towatch, SIGUSR2);
	sigaddset(&sigs_towatch, SIGPIPE);
	sigaddset(&sigs_towatch, SIGHUP);

        /* We're ready */
        struct sembuf sops;
        sops.sem_num = 0;
        sops.sem_op = 1;
        sops.sem_flg = SEM_UNDO;
        if (semop(quit_th_semid, &sops, 1) < 0) {
                perror("semop");
                exit(ERROR);
        }

        while (true) {
		if (sigwait(&sigs_towatch, &sig_caught) != 0) {
			perror("sigwait");
			exit(ERROR);
		}

                switch (sig_caught) {
                case SIGABRT:
			exit(ERROR);
			break;
                case SIGQUIT:
                case SIGTERM:
                case SIGINT:
			exit(main_ret);
                        break;
                case SIGUSR1:
                case SIGUSR2:
                case SIGPIPE:
                case SIGHUP:
                default:
		{
			std::ostringstream msg;

			msg << "Got signal: " << sig_caught;
			mai->sl.log(liblot::logging::PRIORITY(LOG_NOTICE),
				    msg.str());
                       	continue;
		}
		}
        }

        return(0);
}

void dkim_vrfy::args::opt_H::handle(const liblot::args::option &opt)
{
	main_app::instance()->print_title();
	main_app::instance()->print_help();

	exit_main(0);
}

void dkim_vrfy::args::opt_P::handle(const liblot::args::option &opt)
{
	main_app::instance()->set_profile(opt.get_optarg());
}

void dkim_vrfy::args::opt_V::handle(const liblot::args::option &opt)
{
	main_app::instance()->print_title();

	exit_main(0);
}

dkim_vrfy::main_app *dkim_vrfy::main_app::pinstance = 0;

dkim_vrfy::main_app::main_app() : cfga(0)
{
	using namespace liblot::logging;

	cfga = dkim_vrfy::config::new_cfg_access();

	sl.set_ident_current(SYSLOG_NAME);
	sl.open(OPTION(LOG_PID), FACILITY(LOG_MAIL));
}

dkim_vrfy::main_app *dkim_vrfy::main_app::instance() {
	if (pinstance == 0) {
		pinstance = new main_app;
	}
		
	return(pinstance);
}

std::string dkim_vrfy::main_app::get_cfg_value(const std::string &param) const
{
	using namespace liblot::exception;

	std::map<std::string, std::string> &parms_vals = cfga->get_cfg();

	std::map<std::string, std::string>::
		const_iterator mi = parms_vals.find(param);

	if (mi == parms_vals.end()) {
		throw app_notice("Parameter not found");
	}

	return(mi->second);
}

void dkim_vrfy::main_app::print_title()
{
	std::cout << SYSLOG_NAME << " " << VERSION;
	std::cout <<  " Copyright © " << COPYRIGHT_YEAR;
	std::cout << " " << AUTHOR << " " << "<";
	std::cout << AUTHOR_EMAIL << ">" << std::endl;

	std::cout << "License: " << LICENSE << std::endl;
}

void dkim_vrfy::main_app::print_help()
{
	std::cout << "Usage: " << SYSLOG_NAME;
	std::cout << " -P <profile> [-V|-H]\n" << std::endl;
	std::cout << "\t-P <profile>\tSelect configuration profile";
	std::cout << std::endl;
	std::cout << "\t-V\t\tDisplay version & copyright information";
	std::cout << std::endl;
	std::cout << "\t-H\t\tDisplay this help message" << std::endl;
}

int main(int argc, char **argv)
{
	using namespace dkim_vrfy;
	using namespace dkim_vrfy::args;
	using namespace dkim_vrfy::libdkim;
	using namespace liblot::exception;
	using namespace liblot::logging;
	using namespace liblot::cmd;
	using namespace liblot::args;

	main_app *mai = 0;
	try {
		mai = main_app::instance();
	} catch (std::bad_alloc &e) {
		std::cerr << e.what() << std::endl;
		return(ERROR);
	} catch (std::exception &e) {
		std::cerr << e.what() << std::endl;
		return(ERROR);
	} catch (...) {
		std::cerr << "Failed to initialize" << std::endl;
		return(ERROR);
	}

	atexit(dkim_vrfy_atexit);

	sigset_t sigs_toblock;

	sigemptyset(&sigs_toblock);
	sigaddset(&sigs_toblock, SIGABRT);
	sigaddset(&sigs_toblock, SIGTERM);
	sigaddset(&sigs_toblock, SIGQUIT);
	sigaddset(&sigs_toblock, SIGINT);
	sigaddset(&sigs_toblock, SIGUSR1);
	sigaddset(&sigs_toblock, SIGUSR2);
	sigaddset(&sigs_toblock, SIGPIPE);
	sigaddset(&sigs_toblock, SIGHUP);
	sigprocmask(SIG_BLOCK, &sigs_toblock, 0);

	quit_th_semid = semget(IPC_PRIVATE, 1, IPC_CREAT | 0660);
	if (quit_th_semid < 0) {
		perror("semget");
		return(ERROR);
	}

	union semun arg;
	arg.val = 0;
	if (semctl(quit_th_semid, 0, SETVAL, arg) < 0) {
		perror("main semctl");
		return(ERROR);
	}

	if (pthread_create(&th, 0, quit_th, mai) != 0) {
		perror("pthread_create");
		return(ERROR);
	}

	struct sembuf sops;
	sops.sem_num = 0;
	sops.sem_op = -1;
	sops.sem_flg = SEM_UNDO;
	if (semop(quit_th_semid, &sops, 1) < 0) {
		perror("semop");
		exit_main(ERROR);
	}

	std::list<std::string> passthrough_args;

	try {
		/* Argument handling code.  We can take three that, hopefully,
		 * aren't going to be used by the program that we're calling.
		 * In the future, we can make this configurable in the XML.
		 * We should be able to handle command-lines like this in which
		 * order matters and the args have values:
		 * procmail -Y -a $h -d $u
		*/

		getlopt gl(argc, argv);

		option optH(false);
		opt_H optH_handler;
		optH.set_handler(&optH_handler);
		gl.register_opt("-H", &optH);

		option optP(true);
		opt_P optP_handler;
		optP.set_handler(&optP_handler);
		gl.register_opt("-P", &optP);

		option optV(false);
		opt_V optV_handler;
		optV.set_handler(&optV_handler);
		gl.register_opt("-V", &optV);

		opt_nf optnf_handler(&passthrough_args);
		gl.register_nf_handler(&optnf_handler);

		gl.process_args();

		if (mai->get_profile().empty()) {
			std::cerr << "Error: Profile not specified";
			std::cerr << std::endl;
			mai->print_help();
			exit_main(ERROR);
		}

		mai->get_cfga()->read_cfg();
	} catch (app_exception &e) {
		std::cerr << e << std::endl;
		exit_main(ERROR);
	} catch (std::exception &e) {
		std::cerr << e.what() << std::endl;
		exit_main(ERROR);
	} catch (...) {
		std::cerr << "An error occurred" << std::endl;
		exit_main(ERROR);
	}

	exec_cmd *prog = 0;
	try {
		dkim_helper dh(mai->SYSLOG_NAME);
		cmd_parms parms(mai->get_cfg_value("mda"), passthrough_args);
		cmd_opts opts(cmd_opts::OPT_INPUT);
		
		prog = new_exec_cmd(EXEC_IMPL_PIPE, parms, opts);

		std::cin >> dh;

		try {
			dkim_stat err(dh.msg_end());
			dkim_sig dks(dh);

			/* Should we process this message?  We'll pass if
			 * the message was verified by another process
			 * (e.g. dkim-milter) on this server */
			if (dh.check_handle_msg(dks)) {
				std::ostringstream logmsg;
				logmsg << "DKIM processing complete. Status: ";
				logmsg << err;
				
				mai->sl.log(PRIORITY(LOG_NOTICE),
					     logmsg.str());
			
				try {
					/* Don't let any exception thrown at
					 * this point affect delivery of the
					 * message.  If the header isn't
					 * inserted due to an error, that's
					 * OK. */
					std::ostringstream logmsg;
				       
					logmsg << "Inserting header: ";
					logmsg << dh.auth_result_insert(dks);

					mai->sl.log(PRIORITY(LOG_NOTICE),
						    logmsg.str());
				} catch (app_exception &e) {
					mai->sl.log(PRIORITY(LOG_ERR),
						     e.get_str());
				} catch (std::exception &e) {
					mai->sl.log(PRIORITY(LOG_ERR),
						     e.what());
				} catch (...) {
					mai->sl.log(PRIORITY(LOG_ERR), "An unhandled error occurred");
				}
			}
		} catch (app_exception &e) {
			mai->sl.log(PRIORITY(LOG_ERR), e.get_str());
		} catch (std::exception &e) {
			mai->sl.log(PRIORITY(LOG_ERR), e.what());
		} catch (...) {
			mai->sl.log(PRIORITY(LOG_ERR),
				     "An unhandled error occurred");
		}

		std::ostringstream dh_output;

		dh_output << dh;

		prog->write_to_cmd(dh_output.str());

		prog->exec();

		delete prog;
		prog = 0;
	} catch (app_exception &e) {
		mai->sl.log(PRIORITY(LOG_ERR), e.get_str());

		delete prog;
		prog = 0;

		exit_main(ERROR);
	} catch (std::exception &e) {
		mai->sl.log(PRIORITY(LOG_ERR), e.what());

		delete prog;
		prog = 0;

		exit_main(ERROR);
	} catch (...) {
		mai->sl.log(PRIORITY(LOG_ERR),
			     "An unhandled error occurred");

		delete prog;
		prog = 0;

		exit_main(ERROR);
	}

	exit_main(0);

	return(0); // not reached
}
