/* -*- c++ -*- */
/* 
 * This file is part of liblot
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

/* Note: Exceptions aren't used in this module since there is heavy
 * use of Unix/C APIs and C-style error handling (e.g. RAII w/goto);
 * C-style error/status codes are used instead */

#include <iostream>
#include <string>
#include <sstream>
#include <list>
#include <cerrno>
#include <cstring>
#include <cstdio>
#include <cstdlib>

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "liblot_cmd.h"
#ifdef LIBLOT_PROFILE
#include "liblot_profile.h"
#endif

liblot::cmd::exec_cmd *
liblot::cmd::new_exec_cmd(liblot::cmd::exec_impl i,
			  const liblot::cmd::cmd_parms &parms,
			  const liblot::cmd::cmd_opts &opts)
{
	exec_cmd *cmd = 0;

	switch (i) {
	case EXEC_IMPL_PIPE:
		cmd = new exec_cmd_pipe(parms, opts);
		break;
	case EXEC_IMPL_FILE:
		cmd = new exec_cmd_file(parms, opts);
		break;
	default:
		cmd = 0;
	}

	return(cmd);
}

void liblot::cmd::cmd_parms::c_args_init()
{
	const int array_sz = args.size() + 2;

	try {	
		c_args = new char * [array_sz];
	} catch (std::bad_alloc &e) {
		std::cerr << e.what() << std::endl;
		c_args = 0;
		return;
	}

	c_args[0] = strdup(cmd.c_str());

	typedef std::list<std::string>::iterator I;
	int i = 1;
	for (I vi = args.begin(); vi != args.end(); ++vi) {
		c_args[i++] = strdup(vi->c_str());
	}

	c_args[array_sz - 1] = 0;
}

void liblot::cmd::cmd_parms::c_args_free()
{
	if (c_args) {
		for (char **tmp = c_args; *tmp; tmp++) free(*tmp);

		delete[] c_args;
		c_args = 0;
	}
}

liblot::cmd::cmd_parms::cmd_parms &
liblot::cmd::cmd_parms::operator=(const cmd_parms &rhs)
{
	if (this != &rhs) {
		c_args_free();

		cmd = rhs.cmd;
		args = rhs.args;
		c_args_init();
	}

	return(*this);
}

const int liblot::cmd::cmd_opts::OPT_NOWAIT = 1<<0;
const int liblot::cmd::cmd_opts::OPT_OUTPUT = 1<<1;
const int liblot::cmd::cmd_opts::OPT_INPUT = 1<<2;

int liblot::cmd::exec_cmd_pipe::handle_io()
{
	int flags = opts.get_opts();
	bool outpipe_en = flags & cmd_opts::OPT_OUTPUT;
	bool inpipe_en = flags & cmd_opts::OPT_INPUT;
	int ret;

	if (inpipe_en) {
		close(inpipe[0]);
		ret = write_to_pipe(inpipe[1]);
		if (ret == ERROR) return(ERROR);
	}
	if (outpipe_en) {
		close(outpipe[1]);
		ret = read_from_pipe(outpipe[0]);
		if (ret == ERROR) return(ERROR);
	}

	return(0);
}

int liblot::cmd::exec_cmd_pipe::write_to_pipe(int fd)
{
	char *buffer = 0;
	std::streampos sz;
	int local_errno = 0;

	ios_input.seekg (0, std::ios::end);
	sz = ios_input.tellg();
	ios_input.seekg (0, std::ios::beg);

	if (sz == 0) goto cleanup;

	try {	
		buffer = new char [sz];
	} catch (std::bad_alloc &e) {
		std::cerr << e.what() << std::endl;
		close(fd);
		return(ERROR);
	}

	/* Read from ios_input and write to stream here */
        if (ios_input.good()) {
		ios_input.read(buffer, sz);

		ssize_t ret = write(fd, buffer, sz);
		if (ret != sz) {
			local_errno = errno;
			goto cleanup;
		}
	}

cleanup:
	close(fd);
	delete[] buffer;

	if (local_errno) {
		std::cerr << __func__ << " " << strerror(local_errno);
		std::cerr << std::endl;
		return(ERROR);
	}

	return(0);
}

int liblot::cmd::exec_cmd_pipe::read_from_pipe(int fd)
{
	char buf[1024];
	ssize_t nbytes;

	while (os_output.good()) {
		nbytes = read(fd, buf, sizeof(buf));
		if (nbytes == -1) {
			perror("read_from_pipe read");
			return(ERROR);
			break;
		} else if (nbytes == 0) {
			break; // EOF
		} else {
			os_output.write(buf, nbytes);
		}
	}

	return(0);
}

int liblot::cmd::exec_cmd_pipe::exec()
{
	using namespace liblot::cmd;
        pid_t pid;
	int flags = opts.get_opts();
	bool outpipe_en = flags & cmd_opts::OPT_OUTPUT;
	bool inpipe_en = flags & cmd_opts::OPT_INPUT;
	int ret;

	if (inpipe_en) {
		ret = pipe(inpipe);
		if (ret == -1) {
			perror("pipe");
			return(ERROR);
		}
	}
	if (outpipe_en) {
		ret = pipe(outpipe);
		if (ret == -1) {
			perror("pipe");
			return(ERROR);
		}
	}

        pid = fork();
	if (pid == -1) {
		perror("fork");
		return(ERROR);
        } else if (pid == 0) {
                /* Child process */
		if (inpipe_en) {
			ret = close(inpipe[1]); // close writing end
			if (ret == -1) {
				perror("close");
				return(ERROR);
			}
			ret = close(STDIN_FILENO);
			if (ret == -1) {
				perror("close");
				return(ERROR);
			}
			ret = dup2(inpipe[0], STDIN_FILENO);
			if (ret == -1) {
				perror("dup2");
				return(ERROR);
			}
		}
		if (outpipe_en) {
			ret = close(outpipe[0]); // close reading end
			if (ret == -1) {
				perror("close");
				return(ERROR);
			}
			ret = close(STDOUT_FILENO);
			if (ret == -1) {
				perror("close");
				return(ERROR);
			}
			ret = dup2(outpipe[1], STDOUT_FILENO);
			if (ret == -1) {
				perror("dup2");
				return(ERROR);
			}
		}

                execv(parms.get_cmd().c_str(), parms.get_args_array());
                /* If exec fails... */
		perror("execv");
        } else {
                /* Parent process */
		const int waitpid_opts =
			(flags & cmd_opts::OPT_NOWAIT) ? WNOHANG : 0;
        	int status;

		if (inpipe_en || outpipe_en) {
			ret = handle_io();
			if (ret == ERROR) {
				return(ERROR);
			}
		}

                ret = waitpid(pid, &status, waitpid_opts);
		if (ret == -1) {
			perror("waitpid");
			return(ERROR);
		}

                if (WIFEXITED(status)) {
                        return(WEXITSTATUS(status));
                }
        }

        return(ERROR);
}

int liblot::cmd::exec_cmd_file::write_to_file(int fd)
{
	char *buffer = 0;
	std::streampos sz;
	int local_errno = 0;

	ios_input.seekg (0, std::ios::end);
	sz = ios_input.tellg();
	ios_input.seekg (0, std::ios::beg);

	if (sz == 0) goto cleanup;

	try {	
		buffer = new char [sz];
	} catch (std::bad_alloc &e) {
		std::cerr << e.what() << std::endl;
		close(fd);
		return(ERROR);
	}

	/* Read from ios_input and write to stream here */
        if (ios_input.good()) {
		ios_input.read(buffer, sz);

		ssize_t ret = write(fd, buffer, sz);
		if (ret != sz) {
			local_errno = errno;
			goto cleanup;
		}
	}

cleanup:
	delete[] buffer;

	if (local_errno) {
		std::cerr << __func__ << " " << strerror(local_errno);
		std::cerr << std::endl;
		return(ERROR);
	}

	return(0);
}

int liblot::cmd::exec_cmd_file::read_from_file(int fd)
{
	char buf[1024];
	ssize_t nbytes;

	while (os_output.good()) {
		nbytes = read(fd, buf, sizeof(buf));
		if (nbytes == -1) {
			perror("read_from_pipe read");
			return(ERROR);
			break;
		} else if (nbytes == 0) {
			break; // EOF
		} else {
			os_output.write(buf, nbytes);
		}
	}

	return(0);
}

std::string
liblot::cmd::exec_cmd_file::cmdline_generate(const char *in, const char *out)
{
	std::ostringstream os;

	os << parms.get_cmd();
	
	for (std::list<std::string>::const_iterator li = parms.args.begin();
	     li != parms.args.end(); ++li) {
		os << " " << *li;
	}

	if (in) os << " < " << in;
	if (out) os << " > " << out;

	return(os.str());
}

int liblot::cmd::exec_cmd_file::exec()
{
	using namespace liblot::cmd;

	int flags = opts.get_opts();
	bool outpipe_en = flags & cmd_opts::OPT_OUTPUT;
	bool inpipe_en = flags & cmd_opts::OPT_INPUT;
	int ret = 0;
	char in_filename[] = "LIBLOT_IN_XXXXXX";
	char out_filename[] = "LIBLOT_OUT_XXXXXX";
#ifdef LIBLOT_PROFILE
	liblot::profile::stopwatch sw;

	sw.start();
#endif

	if (inpipe_en) {
		int infd = mkstemp(in_filename);
		if (infd == -1) {
			perror("mkstemp");
			return(ERROR);
		}

		ret = write_to_file(infd);
		if (ret == ERROR) goto cleanup;
		
		ret = close(infd);
		if (ret == ERROR) goto cleanup;
	}

	if (outpipe_en) {
		if (mktemp(out_filename) == NULL) {
			perror("mktemp");
			ret = ERROR;
			goto cleanup;
		}
	}

#ifdef LIBLOT_PROFILE
	std::cerr << __func__ << " elapsed: " << sw.look() << std::endl;
#endif
	
	ret = system(cmdline_generate(in_filename, out_filename).c_str());
	if (ret == -1) {
		perror("system");
		ret = ERROR;
		goto cleanup;
	}

#ifdef LIBLOT_PROFILE
	std::cerr << __func__ << " elapsed: " << sw.look() << std::endl;
#endif

	if (outpipe_en) {
		int outfd = open(out_filename, O_RDONLY);
		if (outfd == -1) {
			ret = ERROR;
			goto cleanup;
		}

#ifdef LIBLOT_PROFILE
	std::cerr << __func__ << " elapsed: " << sw.look() << std::endl;
#endif

		ret = read_from_file(outfd);
		if (ret == ERROR) goto cleanup;

#ifdef LIBLOT_PROFILE
	std::cerr << __func__ << " elapsed: " << sw.look() << std::endl;
#endif

		ret = close(outfd);
		if (ret == ERROR) goto cleanup;
	}

#ifdef LIBLOT_PROFILE
	std::cerr << __func__ << " elapsed: " << sw.look() << std::endl;
	sw.stop();
#endif

cleanup:
	unlink(in_filename);
	unlink(out_filename);

        return(ret);
}
