/* -*- c++ -*- */
/* 
 * This file is part of liblot
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef LIBLOT_CMD_H
#define LIBLOT_CMD_H 1

#include <iostream>
#include <sstream>
#include <string>
#include <list>

#ifndef ERROR
#define ERROR -1
#endif

namespace liblot {
	namespace cmd {
		class cmd_parms;
		class cmd_opts;
		class exec_cmd;
		class exec_cmd_pipe;
		class exec_cmd_file;

		enum exec_impl {
			EXEC_IMPL_PIPE,
			EXEC_IMPL_FILE,
		};
		exec_cmd *new_exec_cmd(exec_impl,
				       const cmd_parms &,
				       const cmd_opts &);
	}
}

/* Class containing parameters for exec_cmd (e.g. application
 * name/filepath, command-line args, etc.) */
class liblot::cmd::cmd_parms {
	std::string cmd;
	char **c_args;

	void c_args_init();
	void c_args_free();
public:
	std::list<std::string> args;

	cmd_parms(const std::string &c, const std::list<std::string> &a) :
		cmd(c), c_args(0), args(a) {
		c_args_init();
	}
	~cmd_parms() { 	c_args_free(); }
	cmd_parms(const cmd_parms &rhs) :
		cmd(rhs.cmd), c_args(0), args(rhs.args) {
		c_args_init();
	}
	cmd_parms &operator=(const cmd_parms &);

	std::string get_cmd() const { return(cmd); }
	char * const *get_args_array() const { return(c_args); }
};

class liblot::cmd::cmd_opts {
	int s;
public:
	static const int OPT_NOWAIT;
	static const int OPT_OUTPUT;
	static const int OPT_INPUT;

	cmd_opts(int opts) : s(opts) {}
	cmd_opts() : s(0) {}

	int get_opts() const { return s; }
};

/* Class to facilitate execution of programs (i.e. system()) and
 * management of I/O, if necessary */
class liblot::cmd::exec_cmd {
public:
	virtual ~exec_cmd() {}

	virtual void write_to_cmd(const std::string &) = 0;
	virtual int exec() = 0;
	virtual std::string read_from_cmd() const = 0;
};

class liblot::cmd::exec_cmd_pipe : public exec_cmd {
	cmd_parms parms;
	cmd_opts opts;

	int inpipe[2];
	int outpipe[2];

	std::ostringstream os_output;
	std::stringstream ios_input;

	int handle_io();
	int write_to_pipe(int);
	int read_from_pipe(int);

	/* Don't allow copies */
	exec_cmd_pipe(const exec_cmd_pipe &);
	exec_cmd_pipe &operator=(const exec_cmd_pipe &);
public:
	exec_cmd_pipe(const cmd_parms &p, const cmd_opts &o) :
			parms(p), opts(o),
			os_output(std::ostringstream::binary),
			ios_input(std::istringstream::binary |
			       std::stringstream::in |
			       std::stringstream::out) {}
	
	void write_to_cmd(const std::string &input_data) {
		ios_input << input_data;
	}
	int exec();
	std::string read_from_cmd() const {
		return(os_output.str());
	}
};

class liblot::cmd::exec_cmd_file : public exec_cmd {
	cmd_parms parms;
	cmd_opts opts;

	std::ostringstream os_output;
	std::stringstream ios_input;

	int write_to_file(int);
	int read_from_file(int);
	std::string cmdline_generate(const char *, const char *);

	/* Don't allow copies */
	exec_cmd_file(const exec_cmd_file &);
	exec_cmd_file &operator=(const exec_cmd_file &);
public:
	exec_cmd_file(const cmd_parms &p, const cmd_opts &o) :
			parms(p), opts(o),
			os_output(std::ostringstream::binary),
			ios_input(std::istringstream::binary |
			       std::stringstream::in |
			       std::stringstream::out) {}
	
	void write_to_cmd(const std::string &input_data) {
		ios_input << input_data;
	}
	int exec();
	std::string read_from_cmd() const {
		return(os_output.str());
	}
};
#endif // LIBLOT_CMD_H
