/* -*- c++ -*- */
/* 
 * This file is part of liblot
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef LIBLOT_PROFILE_H
#define LIBLOT_PROFILE_H 1

#include <string>
#include <deque>

#include "liblot_exception.h"

#ifndef ERROR
#define ERROR -1
#endif

namespace liblot {
	namespace profile {
		class stopwatch_result;
		class stopwatch;
	}
}

struct liblot::profile::stopwatch_result {
	int sequence;
	long elapsed;
	std::string comment;
};

/* This class models a stopwatch with an unlimited
 * number of "split" times.  The "buttons" of
 * start/stop and split/reset are modelled here.
 * As in a real stopwatch, pressing "reset" while
 * the clock is running or split when the
 * stopwatch is stopped have no effect.
 * The look() method shows you the current value
 * that would be shown on the face of the stopwatch,
 * and the reset method shows the last splits in
 * sequence. */
class liblot::profile::stopwatch {
	static const std::string PROC_UPTIME_FILE;

	std::deque<long> splits;
	bool running;

	double get_uptime();
	long get_uptime_ms();
public:
	stopwatch() : running(false) {}

	void start() {
		if (! splits.empty()) splits.clear();
		running = true;
		split();
	}
	void split() {
		if (! running) return;

		splits.push_back(get_uptime_ms());
	}
	void stop() {
		if (running) split();
		running = false;
	}

	long look() {
		if (running) return(get_uptime_ms() - splits.front());

		return(splits.front());
	}
	long reset() {
		if (running) return(ERROR);
		if (! splits.empty()) {
			long last_val = splits.front();
			splits.pop_front();

			return(last_val - splits.front());
		} 

		return(0);
	}
};

#endif // LIBLOT_PROFILE_H
