/* -*- c++ -*- */
/* 
 * This file is part of liblot
 * Copyright (C) 2008 Erik Lotspeich <erik@lotspeich.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation: version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef LIBLOT_TEMPLATES_H
#define LIBLOT_TEMPLATES_H 1

#include <map>
#include <queue>
#include <list>

namespace liblot {
	namespace templates {
		template <typename T1, typename T2> class fifo_map;
		template <typename T1, typename T2> class ordered_map;
	}
}

template <typename T1, typename T2>
class liblot::templates::fifo_map {
        std::map<T1, T2> m;
        std::queue<T1> index;
public:
        size_t size() { return(index.size()); }
        bool empty() { return(index.empty()); }
        void push(T1 key, T2 value) {
                m[key] = value;
                index.push(key);
        }
        void pop() {
                m.erase(index.front());
                index.pop();
        }
        T1 front_key() { return(index.front()); }
        T2 front_value() { return(m[index.front()]); }
        T1 back_key() { return(index.back()); }
        T2 back_value() { return(m[index.back()]); }
};

template <typename T1, typename T2>
class liblot::templates::ordered_map {
	struct value_info {
		T2 val;
		typename std::list<T1>::iterator key;
	};
	std::map<T1, value_info> m;
	std::list<T1> index;
public:
	T2 &operator[](const T1 &key) {
		return(m[key].val);
	}

	size_t size() const { return(index.size()); }
	bool empty() const { return(index.empty()); }
	void push(const std::pair<const T1, T2> &elem) {
		push(elem.first, elem.second);
	}
	void push(const T1 &key, const T2 &value) {
		index.push_back(key);

		value_info vinfo;
		vinfo.val = value;
		vinfo.key = --index.end();

		m[key] = vinfo;
	}
	void pop() {
		m.erase(index.front());
		index.pop_front();
	}
	T1 &front_key() { return(index.front()); }
	T2 &front_value() { return(m[index.front()]).val; }
	T1 &back_key() { return(index.back()); }
	T2 &back_value() { return(m[index.back()]).val; }

        class iterator :
		public std::iterator<std::bidirectional_iterator_tag,
               			     std::pair<const T1, value_info> > {
		ordered_map<T1, T2> *om;
		typename std::list<T1>::iterator pos;
		bool om_allocated;
        public:
		iterator() : om(new ordered_map<T1, T2>),
			     pos(om->index.begin()), om_allocated(true) {}
		iterator(const iterator &rhs) :
			om(rhs.om), pos(rhs.pos), om_allocated(false) {}
		iterator(ordered_map<T1, T2> *o,
			 const typename std::list<T1>::iterator &p) :
			om(o), pos(p), om_allocated(false) {}
                ~iterator() { if (om_allocated) delete om; }

		iterator &operator=(const iterator &rhs) {
        		if (this != &rhs) { // avoid self-assignment
				om = rhs.om;
				pos = rhs.pos;	
        		}

        		return(*this);
		}
                friend bool operator==(const iterator &lhs, const iterator &rhs) {
			return((lhs.om == rhs.om) && (lhs.pos == rhs.pos));
                }
                friend bool operator!=(const iterator &lhs, const iterator &rhs) {
                        return(! (lhs == rhs));
                }
                iterator &operator++() {
			pos++;

                        return(*this);
                }
                iterator operator++(int) {
                        iterator tmp(*this);

                        ++(*this);

                        return(*this);
                }
                iterator &operator--() {
			pos--;

                        return(*this);
                }
                iterator operator--(int) {
                        iterator tmp(*this);

                        --(*this);

                        return(*this);
                }
                std::pair<const T1, value_info> &operator*() {
			typename std::map<T1, value_info>::iterator mi = 
					om->m.find(*pos);

			return(*mi);
		}
                std::pair<const T1, value_info> *operator->() {
			typename std::map<T1, value_info>::iterator mi = 
					om->m.find(*pos);

			return(&*mi);
		}
        };

        iterator begin() { return(iterator(this, index.begin())); }
        iterator end() { return(iterator(this, index.end())); }
	iterator find(const T1 &key) {
		typename std::map<T1, value_info>::iterator mi =
				m.find(key);
		if (mi != m.end()) return(iterator(this, mi->second.key));

		return(end());
	}

        class const_iterator :
                public std::iterator<std::bidirectional_iterator_tag,
                                     std::pair<const T1, value_info> > {
		const ordered_map<T1, T2> *om;
		typename std::list<T1>::const_iterator pos;
		bool om_allocated;
        public:
		const_iterator() : om(new ordered_map<T1, T2>),
	     			      pos(om->index.begin()),
				      om_allocated(true) {}
		const_iterator(const const_iterator &rhs) :
			om(rhs.om), pos(rhs.pos), om_allocated(false) {}
		const_iterator(const ordered_map<T1, T2> *o,
			       const typename std::list<T1>::const_iterator &p) :
			om(o), pos(p), om_allocated(false) {}
                ~const_iterator() { if (om_allocated) delete om; }

		const_iterator &operator=(const const_iterator &rhs) {
        		if (this != &rhs) { // avoid self-assignment
				om = rhs.om;
				pos = rhs.pos;	
        		}

        		return(*this);
		}

                friend bool operator==(const const_iterator &lhs,
				       const const_iterator &rhs) {
			return((lhs.om == rhs.om) && (lhs.pos == rhs.pos));
                }
                friend bool operator!=(const const_iterator &lhs,
				       const const_iterator &rhs) {
                        return(! (lhs == rhs));
                }
                const_iterator &operator++() {
			pos++;

                        return(*this);
                }
                const_iterator operator++(int) {
                        const_iterator tmp(*this);

                        ++(*this);

                        return(*this);
                }
                const_iterator &operator--() {
			pos--;

                        return(*this);
                }
                const_iterator operator--(int) {
                        const_iterator tmp(*this);

                        --(*this);

                        return(*this);
                }
                std::pair<const T1, value_info> operator*() const {
			typename std::map<T1, value_info>::const_iterator mi = 
					om->m.find(*pos);

			return(*mi);
		}
                const std::pair<const T1, value_info> *operator->() const {
			typename std::map<T1, value_info>::const_iterator mi =
					om->m.find(*pos);	

			return(&*mi);
		}
        };

        const_iterator begin() const {
		return(const_iterator(this, index.begin()));
	}
        const_iterator end() const {
                return(const_iterator(this, index.end()));
        }
	const_iterator find(const T1 &key) const {
		typename std::map<T1, value_info>::const_iterator mi =
				m.find(key);
		if (mi != m.end()) return(const_iterator(this, mi->second.key));

		return(end());
	}

private:
	void insert(const T1 &key,
		    const typename std::pair<T1, T2> &new_entry,
		    bool next) {
		typename std::map<T1, value_info>::iterator mi =
					m.find(key);
		if (mi != m.end()) {
			value_info vinfo;

			vinfo.val = new_entry.second;
			vinfo.key = index.insert(next ? ++(mi->second.key) :
						 mi->second.key,
						 new_entry.first);

			m[new_entry.first] = vinfo;
		}
	}

public:
	void insert_before(const T1 &key,
			   const typename std::pair<T1, T2> &new_entry) {
		insert(key, new_entry, false);
	}
	void insert_after(const T1 &key,
			   const typename std::pair<T1, T2> &new_entry) {
		insert(key, new_entry, true);
	}
};

#endif // LIBLOT_TEMPLATES_H
